@@ -1,23 +1,22 @@
 /**************************************************************************
- *   browser.c                                                            *
+ *   browser.c  --  This file is part of GNU nano.                        *
  *                                                                        *
  *   Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,  *
  *   2010, 2011, 2013, 2014, 2015 Free Software Foundation, Inc.          *
+ *   Copyright (C) 2015, 2016 Benno Schulenberg                           *
  *                                                                        *
- *   This program is free software; you can redistribute it and/or modify *
- *   it under the terms of the GNU General Public License as published by *
- *   the Free Software Foundation; either version 3, or (at your option)  *
- *   any later version.                                                   *
+ *   GNU nano is free software: you can redistribute it and/or modify     *
+ *   it under the terms of the GNU General Public License as published    *
+ *   by the Free Software Foundation, either version 3 of the License,    *
+ *   or (at your option) any later version.                               *
  *                                                                        *
- *   This program is distributed in the hope that it will be useful, but  *
- *   WITHOUT ANY WARRANTY; without even the implied warranty of           *
- *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    *
- *   General Public License for more details.                             *
+ *   GNU nano is distributed in the hope that it will be useful,          *
+ *   but WITHOUT ANY WARRANTY; without even the implied warranty          *
+ *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.              *
+ *   See the GNU General Public License for more details.                 *
  *                                                                        *
  *   You should have received a copy of the GNU General Public License    *
- *   along with this program; if not, write to the Free Software          *
- *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA            *
- *   02110-1301, USA.                                                     *
+ *   along with this program.  If not, see http://www.gnu.org/licenses/.  *
  *                                                                        *
  **************************************************************************/
 
@@ -36,7 +35,7 @@ static char **filelist = NULL;
 static size_t filelist_len = 0;
 	/* The number of files in the list. */
 static int width = 0;
-	/* The number of files that we can display per line. */
+	/* The number of files that we can display per screen row. */
 static int longest = 0;
 	/* The number of columns in the longest filename in the list. */
 static size_t selected = 0;
@@ -60,7 +59,6 @@ char *do_browser(char *path)
 
     /* Don't show a cursor in the file list. */
     curs_set(0);
-    blank_statusbar();
 
   read_directory_contents:
 	/* We come here when we refresh or select a new directory. */
@@ -71,7 +69,7 @@ char *do_browser(char *path)
 	dir = opendir(path);
 
     if (path == NULL || dir == NULL) {
-	statusline(ALERT, "Cannot open directory: %s", strerror(errno));
+	statusline(ALERT, _("Cannot open directory: %s"), strerror(errno));
 	/* If we don't have a file list yet, there is nothing to show. */
 	if (filelist == NULL) {
 	    napms(1200);
@@ -149,7 +147,7 @@ char *do_browser(char *path)
 		/* If we selected the same filename as last time, fake a
 		 * press of the Enter key so that the file is read in. */
 		if (old_selected == selected)
-		    unget_kbinput(sc_seq_or(do_enter, 0), FALSE, FALSE);
+		    unget_kbinput(KEY_ENTER, FALSE);
 	    }
 
 	    continue;
@@ -221,28 +219,18 @@ char *do_browser(char *path)
 	    selected = filelist_len - 1;
 	} else if (func == goto_dir_void) {
 	    /* Ask for the directory to go to. */
-	    int i = do_prompt(TRUE,
-#ifndef DISABLE_TABCOMP
-			FALSE,
-#endif
-			MGOTODIR, NULL,
+	    int i = do_prompt(TRUE, FALSE, MGOTODIR, NULL,
 #ifndef DISABLE_HISTORIES
 			NULL,
 #endif
 			/* TRANSLATORS: This is a prompt. */
 			browser_refresh, _("Go To Directory"));
 
-	    /* If the directory begins with a newline (i.e. an
-	     * encoded null), treat it as though it's blank. */
-	    if (i < 0 || *answer == '\n') {
+	    if (i < 0) {
 		statusbar(_("Cancelled"));
 		continue;
 	    }
 
-	    /* Convert newlines to nulls in the directory name. */
-	    sunder(answer);
-	    align(&answer);
-
 	    path = free_and_assign(path, real_dir_from_tilde(answer));
 
 	    /* If the given path is relative, join it with the current path. */
@@ -309,7 +297,7 @@ char *do_browser(char *path)
 	    /* If we are moving up one level, remember where we came from, so
 	     * this directory can be highlighted and easily reentered. */
 	    if (strcmp(tail(filelist[selected]), "..") == 0)
-		present_name = striponedir(filelist[selected]);
+		present_name = strip_last_component(filelist[selected]);
 
 	    /* Try opening and reading the selected directory. */
 	    path = mallocstrcpy(path, filelist[selected]);
@@ -356,8 +344,6 @@ char *do_browse_from(const char *inpath)
     char *path;
 	/* This holds the tilde-expanded version of inpath. */
 
-    assert(inpath != NULL);
-
     path = real_dir_from_tilde(inpath);
 
     /* Perhaps path is a directory.  If so, we'll pass it to
@@ -367,7 +353,7 @@ char *do_browse_from(const char *inpath)
      * at all.  If so, we'll just pass the current directory to
      * do_browser(). */
     if (stat(path, &st) == -1 || !S_ISDIR(st.st_mode)) {
-	path = free_and_assign(path, striponedir(path));
+	path = free_and_assign(path, strip_last_component(path));
 
 	if (stat(path, &st) == -1 || !S_ISDIR(st.st_mode)) {
 	    char * currentdir = charalloc(PATH_MAX + 1);
@@ -377,12 +363,11 @@ char *do_browse_from(const char *inpath)
 
 	    if (path == NULL) {
 		free(currentdir);
-		statusline(MILD, "The working directory has disappeared");
+		statusline(MILD, _("The working directory has disappeared"));
 		beep();
 		napms(1200);
 		return NULL;
-	    } else
-		align(&path);
+	    }
 	}
     }
 
@@ -400,7 +385,7 @@ char *do_browse_from(const char *inpath)
  * set filelist_len to the number of files in that list, set longest to
  * the width in columns of the longest filename in that list (between 15
  * and COLS), and set width to the number of files that we can display
- * per line.  And sort the list too. */
+ * per screen row.  And sort the list too. */
 void read_the_list(const char *path, DIR *dir)
 {
     const struct dirent *nextdir;
@@ -482,6 +467,10 @@ functionptrtype parse_browser_input(int *kbinput)
 		return do_help_void;
 	    case 'E':
 	    case 'e':
+	    case 'Q':
+	    case 'q':
+	    case 'X':
+	    case 'x':
 		return do_exit;
 	    case 'G':
 	    case 'g':
@@ -497,13 +486,15 @@ functionptrtype parse_browser_input(int *kbinput)
     return func_from_key(kbinput);
 }
 
-/* Set width to the number of files that we can display per line, if
- * necessary, and display the list of files. */
+/* Set width to the number of files that we can display per screen row,
+ * if necessary, and display the list of files. */
 void browser_refresh(void)
 {
     size_t i;
-    int line = 0, col = 0;
-	/* The current line and column while the list is getting displayed. */
+    int row = 0, col = 0;
+	/* The current row and column while the list is getting displayed. */
+    int the_row = 0, the_column = 0;
+	/* The row and column of the selected item. */
     char *info;
 	/* The additional information that we'll display about a file. */
 
@@ -514,7 +505,7 @@ void browser_refresh(void)
 
     i = selected - selected % (editwinrows * width);
 
-    for (; i < filelist_len && line < editwinrows; i++) {
+    for (; i < filelist_len && row < editwinrows; i++) {
 	struct stat st;
 	const char *thename = tail(filelist[i]);
 		/* The filename we display, minus the path. */
@@ -523,51 +514,45 @@ void browser_refresh(void)
 	size_t infolen;
 		/* The length of the file information in columns. */
 	int infomaxlen = 7;
-		/* The maximum length of the file information in
-		 * columns: seven for "--", "(dir)", or the file size,
-		 * and 12 for "(parent dir)". */
+		/* The maximum length of the file information in columns:
+		 * normally seven, but will be twelve for "(parent dir)". */
 	bool dots = (COLS >= 15 && namelen >= longest - infomaxlen);
-		/* Do we put an ellipsis before the filename?  Don't set
-		 * this to TRUE if we have fewer than 15 columns (i.e.
-		 * one column for padding, plus seven columns for a
-		 * filename other than ".."). */
-	char *disp = display_string(thename, dots ? namelen -
-		longest + infomaxlen + 4 : 0, longest, FALSE);
-		/* If we put an ellipsis before the filename, reserve
-		 * one column for padding, plus seven columns for "--",
-		 * "(dir)", or the file size, plus three columns for the
-		 * ellipsis. */
+		/* Whether to put an ellipsis before the filename?  We don't
+		 * waste space on dots when there are fewer than 15 columns. */
+	char *disp = display_string(thename, dots ?
+		namelen + infomaxlen + 4 - longest : 0, longest, FALSE);
+		/* The filename (or a fragment of it) in displayable format.
+		 * When a fragment, account for dots plus one space padding. */
 
-	/* Start highlighting the currently selected file or directory. */
-	if (i == selected)
+	/* If this is the selected item, start its highlighting, and
+	 * remember its location to be able to place the cursor on it. */
+	if (i == selected) {
 	    wattron(edit, hilite_attribute);
+	    the_row = row;
+	    the_column = col;
+	}
 
-	blank_line(edit, line, col, longest);
+	blank_row(edit, row, col, longest);
 
-	/* If dots is TRUE, we will display something like "...ename". */
+	/* If the name is too long, we display something like "...ename". */
 	if (dots)
-	    mvwaddstr(edit, line, col, "...");
-	mvwaddstr(edit, line, dots ? col + 3 : col, disp);
+	    mvwaddstr(edit, row, col, "...");
+	mvwaddstr(edit, row, dots ? col + 3 : col, disp);
 
 	free(disp);
 
 	col += longest;
 
-	/* Show information about the file.  We don't want to report
-	 * file sizes for links, so we use lstat(). */
+	/* Show information about the file: "--" for symlinks (except when
+	 * they point to a directory) and for files that have disappeared,
+	 * "(dir)" for directories, and the file size for normal files. */
 	if (lstat(filelist[i], &st) == -1 || S_ISLNK(st.st_mode)) {
-	    /* If the file doesn't exist (i.e. it's been deleted while
-	     * the file browser is open), or it's a symlink that doesn't
-	     * point to a directory, display "--". */
 	    if (stat(filelist[i], &st) == -1 || !S_ISDIR(st.st_mode))
 		info = mallocstrcpy(NULL, "--");
-	    /* If the file is a symlink that points to a directory,
-	     * display it as a directory. */
 	    else
 		/* TRANSLATORS: Try to keep this at most 7 characters. */
 		info = mallocstrcpy(NULL, _("(dir)"));
 	} else if (S_ISDIR(st.st_mode)) {
-	    /* If the file is a directory, display it as such. */
 	    if (strcmp(thename, "..") == 0) {
 		/* TRANSLATORS: Try to keep this at most 12 characters. */
 		info = mallocstrcpy(NULL, _("(parent dir)"));
@@ -580,6 +565,7 @@ void browser_refresh(void)
 
 	    info = charalloc(infomaxlen + 1);
 
+	    /* Massage the file size into a human-readable form. */
 	    if (st.st_size < (1 << 10))
 		modifier = ' ';  /* bytes */
 	    else if (st.st_size < (1 << 20)) {
@@ -593,8 +579,7 @@ void browser_refresh(void)
 		modifier = 'G';  /* gigabytes */
 	    }
 
-	    /* Show the size if less than a terabyte,
-	     * otherwise show "(huge)". */
+	    /* Show the size if less than a terabyte, else show "(huge)". */
 	    if (result < (1 << 10))
 		sprintf(info, "%4ju %cB", (intmax_t)result, modifier);
 	    else
@@ -606,13 +591,13 @@ void browser_refresh(void)
 	/* Make sure info takes up no more than infomaxlen columns. */
 	infolen = strlenpt(info);
 	if (infolen > infomaxlen) {
-	    null_at(&info, actual_x(info, infomaxlen));
+	    info[actual_x(info, infomaxlen)] = '\0';
 	    infolen = infomaxlen;
 	}
 
-	mvwaddstr(edit, line, col - infolen, info);
+	mvwaddstr(edit, row, col - infolen, info);
 
-	/* Finish highlighting the currently selected file or directory. */
+	/* If this is the selected item, finish its highlighting. */
 	if (i == selected)
 	    wattroff(edit, hilite_attribute);
 
@@ -621,14 +606,18 @@ void browser_refresh(void)
 	/* Add some space between the columns. */
 	col += 2;
 
-	/* If the next entry isn't going to fit on the current line,
-	 * move to the next line. */
+	/* If the next entry isn't going to fit on the current row,
+	 * move to the next row. */
 	if (col > COLS - longest) {
-	    line++;
+	    row++;
 	    col = 0;
 	}
+    }
 
-	wmove(edit, line, col);
+    /* If requested, put the cursor on the selected item and switch it on. */
+    if (ISSET(SHOW_CURSOR)) {
+	wmove(edit, the_row, the_column);
+	curs_set(1);
     }
 
     wnoutrefresh(edit);
@@ -679,15 +668,11 @@ int filesearch_init(void)
 	buf = mallocstrcpy(NULL, "");
 
     /* This is now one simple call.  It just does a lot. */
-    input = do_prompt(FALSE,
-#ifndef DISABLE_TABCOMP
-	TRUE,
-#endif
-	MWHEREISFILE, NULL,
+    input = do_prompt(FALSE, FALSE, MWHEREISFILE, NULL,
 #ifndef DISABLE_HISTORIES
-	&search_history,
+		&search_history,
 #endif
-	browser_refresh, "%s%s", _("Search"), buf);
+		browser_refresh, "%s%s", _("Search"), buf);
 
     /* Release buf now that we don't need it anymore. */
     free(buf);
@@ -716,7 +701,7 @@ void findnextfile(const char *needle)
     /* Save the settings of all flags. */
     memcpy(stash, flags, sizeof(flags));
 
-    /* Search forward, case insensitive and without regexes. */
+    /* Search forward, case insensitive, and without regexes. */
     UNSET(BACKWARDS_SEARCH);
     UNSET(CASE_SENSITIVE);
     UNSET(USE_REGEXP);
@@ -800,23 +785,17 @@ void do_last_file(void)
     selected = filelist_len - 1;
 }
 
-/* Strip one directory from the end of path, and return the stripped
- * path.  The returned string is dynamically allocated, and should be
- * freed. */
-char *striponedir(const char *path)
+/* Strip one element from the end of path, and return the stripped path.
+ * The returned string is dynamically allocated, and should be freed. */
+char *strip_last_component(const char *path)
 {
-    char *retval, *tmp;
+    char *copy = mallocstrcpy(NULL, path);
+    char *last_slash = strrchr(copy, '/');
 
-    assert(path != NULL);
+    if (last_slash != NULL)
+	*last_slash = '\0';
 
-    retval = mallocstrcpy(NULL, path);
-
-    tmp = strrchr(retval, '/');
-
-    if (tmp != NULL)
-	null_at(&retval, tmp - retval);
-
-    return retval;
+    return copy;
 }
 
 #endif /* !DISABLE_BROWSER */
